/*
 * Copyright (c) 2016-2021 Satlab A/S <satlab@satlab.com>
 * All rights reserved. Do not distribute without permission.
 */

#define PY_SSIZE_T_CLEAN
#include <Python.h>
#include <structmember.h>

#include <prop-client/prop_client.h>

#define PROP_CLIENT_SUBSYS_SPEC_CAPSULE	"propclient.subsys_spec"

/* PropNode object */
typedef struct {
	PyObject_HEAD
	uint8_t node;
	uint8_t port;
	uint32_t timeout;
	const struct prop_subsys_spec *spec;
} PropNodeObject;

static int PropNode_init(PropNodeObject *self, PyObject *args, PyObject *kwds)
{
	static char *kwlist[] = {"node", "port", "spec", "timeout", NULL};
	PyObject *subsys_spec_capsule;

	self->timeout = 1000;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "bbO|I", kwlist, &self->node, &self->port, &subsys_spec_capsule, &self->timeout))
		return -1;

	self->spec = PyCapsule_GetPointer(subsys_spec_capsule, PROP_CLIENT_SUBSYS_SPEC_CAPSULE);
	if (!self->spec)
		return -1;

	return 0;
}

static PyObject *PropNode_get_ex(PropNodeObject *self, PyObject *props_arg, uint32_t timeout)
{
	int ret;
	PyObject *value;
	const struct prop_spec *prop;
	struct sl_prop_query query;
	PyObject *propobj, *props_ascii, *proplist, *propretlist;
	bool unpack_single;

	/* Could be argument, but better to let C interface hide this in the future */
	size_t tx_size = 1024, rx_size = 1024;

	/* Create list of properties to get */
	if (PyList_Check(props_arg)) {
		/* Argument is already list, use direcly */
		proplist = props_arg;

		/* Make borrowed reference into owned */
		Py_INCREF(proplist);

		/* Return as list */
		unpack_single = false;
	} else if (PyUnicode_Check(props_arg)) {
		/* Create single entry list */
		proplist = PyList_New(0);
		if (!proplist)
			goto err;

		if (PyList_Append(proplist, props_arg) != 0)
			goto err_proplist;

		/* Return value object directy */
		unpack_single = true;
	} else {
		PyErr_Format(PyExc_ValueError, "Properties must either be string or list of strings");
		goto err;
	}

	/* Prepare query */
	ret = sl_prop_remote_query_create(&query, tx_size, rx_size);
	if (ret < 0) {
		PyErr_Format(PyExc_MemoryError, "Failed to allocate query");
		goto err_proplist;
	}

	/* Add properties to query */
	for (Py_ssize_t i = 0; i < PyList_Size(proplist); i++) {
		propobj = PyList_GetItem(proplist, i);
		if (!propobj) {
			goto err_query;
		}

		props_ascii = PyUnicode_AsASCIIString(propobj);
		if (!props_ascii) {
			goto err_query;
		}

		const char *name = PyBytes_AsString(props_ascii);
		if (!name) {
			goto err_query;
		}

		prop = sl_prop_spec_find_by_name(self->spec, name);
		if (!prop) {
			PyErr_Format(PyExc_ValueError, "No such property \"%s\"", name);
			goto err_query;
		}

		/* Add to query */
		ret = sl_prop_remote_query_get(&query, prop->id);
		if (ret < 0)
			goto err_query;
	}

	/* Send query */
	Py_BEGIN_ALLOW_THREADS
	ret = sl_prop_remote_query_send(&query, self->node, self->port, timeout);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		PyErr_Format(PyExc_Exception, "Failed to send property query, error %d", ret);
		goto err_query;
	}

	/* Create return list */
	propretlist = PyList_New(PyList_Size(proplist));
	if (!propretlist)
		goto err_query;

	/* Unpack replies */
	for (Py_ssize_t i = 0; i < PyList_Size(proplist); i++) {
		propobj = PyList_GetItem(proplist, i);
		if (!propobj) {
			goto err_propretlist;
		}

		props_ascii = PyUnicode_AsASCIIString(propobj);
		if (!props_ascii) {
			goto err_propretlist;
		}

		const char *name = PyBytes_AsString(props_ascii);
		if (!name) {
			goto err_ascii;
		}

		prop = sl_prop_spec_find_by_name(self->spec, name);
		if (!prop) {
			PyErr_Format(PyExc_ValueError, "No such property \"%s\"", name);
			goto err_ascii;
		}

		/* Create value object */
		switch (prop->type.type) {
		case PROP_TYPE_BOOL: {
			bool boolval;
			ret = sl_prop_remote_query_get_reply(&query, prop->id, &boolval, sizeof(boolval));
			if (ret < 0)
				goto err_ascii;

			value = PyBool_FromLong(boolval);
			break;
		}
		case PROP_TYPE_UINT8: {
			uint8_t u8val;
			ret = sl_prop_remote_query_get_reply(&query, prop->id, &u8val, sizeof(u8val));
			if (ret < 0)
				goto err_ascii;

			value = PyLong_FromLong(u8val);
			break;
		}
		case PROP_TYPE_UINT16: {
			uint16_t u16val;
			ret = sl_prop_remote_query_get_reply(&query, prop->id, &u16val, sizeof(u16val));
			if (ret < 0)
				goto err_ascii;

			value = PyLong_FromLong(u16val);
			break;
		}
		case PROP_TYPE_UINT32: {
			uint32_t u32val;
			ret = sl_prop_remote_query_get_reply(&query, prop->id, &u32val, sizeof(u32val));
			if (ret < 0)
				goto err_ascii;

			value = PyLong_FromLong(u32val);
			break;
		}
		case PROP_TYPE_UINT64: {
			uint64_t u64val;
			ret = sl_prop_remote_query_get_reply(&query, prop->id, &u64val, sizeof(u64val));
			if (ret < 0)
				goto err_ascii;

			value = PyLong_FromLong(u64val);
			break;
		}
		case PROP_TYPE_INT8: {
			int8_t i8val;
			ret = sl_prop_remote_query_get_reply(&query, prop->id, &i8val, sizeof(i8val));
			if (ret < 0)
				goto err_ascii;

			value = PyLong_FromLong(i8val);
			break;
		}
		case PROP_TYPE_INT16: {
			int16_t i16val;
			ret = sl_prop_remote_query_get_reply(&query, prop->id, &i16val, sizeof(i16val));
			if (ret < 0)
				goto err_ascii;

			value = PyLong_FromLong(i16val);
			break;
		}
		case PROP_TYPE_INT32: {
			int32_t i32val;
			ret = sl_prop_remote_query_get_reply(&query, prop->id, &i32val, sizeof(i32val));
			if (ret < 0)
				goto err_ascii;

			value = PyLong_FromLong(i32val);
			break;
		}
		case PROP_TYPE_INT64: {
			int64_t i64val;
			ret = sl_prop_remote_query_get_reply(&query, prop->id, &i64val, sizeof(i64val));
			if (ret < 0)
				goto err_ascii;

			value = PyLong_FromLong(i64val);
			break;
		}
		case PROP_TYPE_DOUBLE: {
			double doubleval;
			ret = sl_prop_remote_query_get_reply(&query, prop->id, &doubleval, sizeof(doubleval));
			if (ret < 0)
				goto err_ascii;

			value = PyFloat_FromDouble(doubleval);
			break;
		}
		case PROP_TYPE_FLOAT: {
			float floatval;
			ret = sl_prop_remote_query_get_reply(&query, prop->id, &floatval, sizeof(floatval));
			if (ret < 0)
				goto err_ascii;

			value = PyFloat_FromDouble(floatval);
			break;
		}
		case PROP_TYPE_STRING: {
			char *string = malloc(prop->size);
			if (!string) {
				PyErr_Format(PyExc_MemoryError, "No available memory for string");
				goto err_ascii;
			}

			ret = sl_prop_remote_query_get_reply(&query, prop->id, string, prop->size);
			if (ret < 0) {
				free(string);
				goto err_ascii;
			}

			string[prop->size - 1] = '\0';
			value = PyUnicode_FromString(string);
			free(string);
			break;
		}
		case PROP_TYPE_BINARY: {
			uint8_t *bin = calloc(1, prop->size);
			if (!bin) {
				PyErr_Format(PyExc_MemoryError, "No available memory for bytes");
				goto err_ascii;
			}

			ret = sl_prop_remote_query_get_reply(&query, prop->id, bin, prop->size);
			if (ret < 0) {
				free(bin);
				goto err_ascii;
			}

			value = PyBytes_FromStringAndSize((const char *)bin, prop->size);
			free(bin);
			break;
		}
		default:
			PyErr_Format(PyExc_Exception, "Unknown type for property \"%s\"", name);
			goto err_ascii;
		}

		/* We no longer need reference to ASCII name */
		Py_DECREF(props_ascii);

		/* Check generated object */
		if (!value)
			goto err_propretlist;

		/* Add value object to return list */
		PyList_SetItem(propretlist, i, value);
	}

	/* Free query memory */
	sl_prop_remote_query_destroy(&query);

	/* Make sure to drop reference to property list */
	Py_DECREF(proplist);

	/* If getting single value, return value directly instead of as list */
	if (unpack_single) {
		value = PyList_GetItem(propretlist, 0);
		Py_INCREF(value);
		Py_DECREF(propretlist);

		return value;
	}

	return propretlist;

err_ascii:
	/* We no longer need reference to ASCII name */
	Py_DECREF(props_ascii);
err_propretlist:
	/* Make sure to drop reference to property return list */
	Py_DECREF(propretlist);
err_query:
	/* Free query memory */
	sl_prop_remote_query_destroy(&query);
err_proplist:
	/* Make sure to drop reference to property list */
	Py_DECREF(proplist);
err:
	return NULL;
}

static PyObject *PropNode_get(PropNodeObject *self, PyObject *args, PyObject *kwds)
{
	/* Arguments */
	static char *kwlist[] = {"props", "timeout", NULL};
	PyObject *props_arg;
	uint32_t timeout = self->timeout;

	/* Parse arguments */
	if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|I", kwlist, &props_arg, &timeout))
		return NULL;

	return PropNode_get_ex(self, props_arg, timeout);
}

static PyObject *PropNode_set_ex(PropNodeObject *self, PyObject *props_arg, PyObject *values_arg, uint32_t timeout)
{
	int ret;
	const struct prop_spec *prop;
	struct sl_prop_query query;
	PyObject *propobj, *props_ascii, *proplist, *valobj, *vallist;

	/* Could be argument, but better to let C interface hide this in the future */
	size_t tx_size = 1024, rx_size = 1024;

	/* Create list of properties to set */
	if (PyList_Check(props_arg)) {
		/* Argument is already list, use direcly */
		proplist = props_arg;

		/* Make borrowed reference into owned */
		Py_INCREF(proplist);
	} else if (PyUnicode_Check(props_arg)) {
		/* Create single entry list */
		proplist = PyList_New(0);
		if (!proplist)
			goto err;

		if (PyList_Append(proplist, props_arg) != 0)
			goto err_proplist;
	} else {
		PyErr_Format(PyExc_ValueError, "Properties must either be string or list of strings");
		goto err;
	}

	/* Create list of values to set */
	if (PyList_Check(values_arg)) {
		/* Argument is already list, use direcly */
		vallist = values_arg;

		/* Make borrowed reference into owned */
		Py_INCREF(vallist);
	} else {
		/* Create single entry list */
		vallist = PyList_New(0);
		if (!vallist)
			goto err;

		if (PyList_Append(vallist, values_arg) != 0)
			goto err_vallist;
	}

	/* Verify key/value count match */
	if (PyList_Size(proplist) != PyList_Size(vallist)) {
		PyErr_Format(PyExc_ValueError, "key/value count does not match");
		goto err_vallist;
	}

	/* Prepare query */
	ret = sl_prop_remote_query_create(&query, tx_size, rx_size);
	if (ret < 0) {
		PyErr_Format(PyExc_MemoryError, "Failed to allocate query");
		goto err_proplist;
	}

	/* Add properties to query */
	for (Py_ssize_t i = 0; i < PyList_Size(proplist); i++) {
		propobj = PyList_GetItem(proplist, i);
		if (!propobj) {
			goto err_query;
		}

		valobj = PyList_GetItem(vallist, i);
		if (!valobj) {
			goto err_query;
		}

		props_ascii = PyUnicode_AsASCIIString(propobj);
		if (!props_ascii) {
			goto err_query;
		}

		const char *name = PyBytes_AsString(props_ascii);
		if (!name) {
			goto err_ascii;
		}

		prop = sl_prop_spec_find_by_name(self->spec, name);
		if (!prop) {
			PyErr_Format(PyExc_ValueError, "No such property \"%s\"", name);
			goto err_ascii;
		}

		switch (prop->type.type) {
		case PROP_TYPE_BOOL: {
			if (!PyBool_Check(valobj)) {
				PyErr_Format(PyExc_ValueError, "Invalid boolean value");
				goto err_ascii;
			}

			bool boolval = (valobj == Py_True);
			ret = sl_prop_remote_query_set(&query, prop->id, &boolval, sizeof(boolval));
			break;
		}
		case PROP_TYPE_UINT8: {
			long longval = PyLong_AsLong(valobj);
			if (PyErr_Occurred() || longval < 0 || longval > UINT8_MAX) {
				PyErr_Format(PyExc_ValueError, "Invalid value for UINT8 property");
				goto err_ascii;
			}

			uint8_t u8val = longval;
			ret = sl_prop_remote_query_set(&query, prop->id, &u8val, sizeof(u8val));
			break;
		}
		case PROP_TYPE_UINT16: {
			long longval = PyLong_AsLong(valobj);
			if (PyErr_Occurred() || longval < 0 || longval > UINT16_MAX) {
				PyErr_Format(PyExc_ValueError, "Invalid value for UINT16 property");
				goto err_ascii;
			}

			uint16_t u16val = longval;
			ret = sl_prop_remote_query_set(&query, prop->id, &u16val, sizeof(u16val));
			break;
		}
		case PROP_TYPE_UINT32: {
			long longval = PyLong_AsLong(valobj);
			if (PyErr_Occurred() || longval < 0 || longval > UINT32_MAX) {
				PyErr_Format(PyExc_ValueError, "Invalid value for UINT32 property");
				goto err_ascii;
			}

			uint32_t u32val = longval;
			ret = sl_prop_remote_query_set(&query, prop->id, &u32val, sizeof(u32val));
			break;
		}
		case PROP_TYPE_UINT64: {
			long longval = PyLong_AsLong(valobj);
			if (PyErr_Occurred() || longval < 0 || longval > UINT64_MAX) {
				PyErr_Format(PyExc_ValueError, "Invalid value for UINT64 property");
				goto err_ascii;
			}

			uint64_t u64val = longval;
			ret = sl_prop_remote_query_set(&query, prop->id, &u64val, sizeof(u64val));
			break;
		}
		case PROP_TYPE_INT8: {
			long longval = PyLong_AsLong(valobj);
			if (PyErr_Occurred() || longval < INT8_MIN || longval > INT8_MAX) {
				PyErr_Format(PyExc_ValueError, "Invalid value for INT8 property");
				goto err_ascii;
			}

			int8_t i8val = longval;
			ret = sl_prop_remote_query_set(&query, prop->id, &i8val, sizeof(i8val));
			break;
		}
		case PROP_TYPE_INT16: {
			long longval = PyLong_AsLong(valobj);
			if (PyErr_Occurred() || longval < INT16_MIN || longval > INT16_MAX) {
				PyErr_Format(PyExc_ValueError, "Invalid value for INT16 property");
				goto err_ascii;
			}

			int16_t i16val = longval;
			ret = sl_prop_remote_query_set(&query, prop->id, &i16val, sizeof(i16val));
			break;
		}
		case PROP_TYPE_INT32: {
			long longval = PyLong_AsLong(valobj);
			if (PyErr_Occurred() || longval < INT32_MIN || longval > INT32_MAX) {
				PyErr_Format(PyExc_ValueError, "Invalid value for INT32 property");
				goto err_ascii;
			}

			int32_t i32val = longval;
			ret = sl_prop_remote_query_set(&query, prop->id, &i32val, sizeof(i32val));
			break;
		}
		case PROP_TYPE_INT64: {
			long longval = PyLong_AsLong(valobj);
			if (PyErr_Occurred() || longval < INT64_MIN || longval > INT64_MAX) {
				PyErr_Format(PyExc_ValueError, "Invalid value for INT64 property");
				goto err_ascii;
			}

			int64_t i64val = longval;
			ret = sl_prop_remote_query_set(&query, prop->id, &i64val, sizeof(i64val));
			break;
		}
		case PROP_TYPE_DOUBLE: {
			double doubleval = PyFloat_AsDouble(valobj);
			if (PyErr_Occurred()) {
				PyErr_Format(PyExc_ValueError, "Invalid value for double property");
				goto err_ascii;
			}

			ret = sl_prop_remote_query_set(&query, prop->id, &doubleval, sizeof(doubleval));
			break;
		}
		case PROP_TYPE_FLOAT: {
			float floatval = PyFloat_AsDouble(valobj);
			if (PyErr_Occurred()) {
				PyErr_Format(PyExc_ValueError, "Invalid value for float property");
				goto err_ascii;
			}

			ret = sl_prop_remote_query_set(&query, prop->id, &floatval, sizeof(floatval));
			break;
		}
		case PROP_TYPE_STRING: {
			PyObject *string_ascii = PyUnicode_AsASCIIString(valobj);
			if (!string_ascii)
				goto err_ascii;

			char *string;
			Py_ssize_t string_size;
			if (PyBytes_AsStringAndSize(string_ascii, &string, &string_size) < 0) {
				Py_DECREF(string_ascii);
				goto err_ascii;
			}

			ret = sl_prop_remote_query_set(&query, prop->id, &string, string_size);
			Py_DECREF(string_ascii);
			break;
		}
		case PROP_TYPE_BINARY: {
			char *bin;
			Py_ssize_t bin_size;
			if (PyBytes_AsStringAndSize(valobj, &bin, &bin_size) < 0)
				goto err_ascii;

			ret = sl_prop_remote_query_set(&query, prop->id, (const uint8_t *)bin, bin_size);
			break;
		}
		default:
			PyErr_Format(PyExc_Exception, "Unknown type for property \"%s\"", name);
			break;
		}

		if (ret < 0) {
			goto err_ascii;
		}
	}

	/* Send query */
	Py_BEGIN_ALLOW_THREADS
	ret = sl_prop_remote_query_send(&query, self->node, self->port, timeout);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		PyErr_Format(PyExc_Exception, "Failed to send property query, error %d", ret);
		goto err_query;
	}

	/* Free query memory */
	sl_prop_remote_query_destroy(&query);

	/* Make sure to drop reference to property list */
	Py_DECREF(proplist);

	Py_RETURN_NONE;

err_ascii:
	/* We no longer need reference to ASCII name */
	Py_DECREF(props_ascii);
err_query:
	/* Free query memory */
	sl_prop_remote_query_destroy(&query);
err_vallist:
	/* Make sure to drop reference to property value list */
	Py_DECREF(vallist);
err_proplist:
	/* Make sure to drop reference to property list */
	Py_DECREF(proplist);
err:
	return NULL;
}

static PyObject *PropNode_set(PropNodeObject *self, PyObject *args, PyObject *kwds)
{
	/* Arguments */
	static char *kwlist[] = {"props", "values", "timeout", NULL};
	PyObject *props_arg, *values_arg;
	uint32_t timeout = self->timeout;

	/* Parse arguments */
	if (!PyArg_ParseTupleAndKeywords(args, kwds, "OO|I", kwlist, &props_arg, &values_arg, &timeout))
		return NULL;

	return PropNode_set_ex(self, props_arg, values_arg, timeout);
}

static PyObject *PropNode_save(PropNodeObject *self, PyObject *args, PyObject *kwds)
{
	int ret;
	const struct prop_group_spec *group;

	static char *kwlist[] = {"name", "fallback", "timeout", NULL};
	const char *name;
	uint32_t timeout = self->timeout;
	int fallback = false;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|pI", kwlist, &name, &fallback, &timeout))
		return NULL;

	group = sl_prop_group_spec_find_by_name(self->spec, name);
	if (!group) {
		PyErr_Format(PyExc_ValueError, "No such property group \"%s\"", name);
		return NULL;
	}

	Py_BEGIN_ALLOW_THREADS
	ret = sl_prop_remote_save(self->node, self->port, timeout, group->id, fallback);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		PyErr_Format(PyExc_Exception, "Failed to save group, error %d", ret);
		return NULL;
	}

	Py_RETURN_NONE;
}

static PyObject *PropNode_load(PropNodeObject *self, PyObject *args, PyObject *kwds)
{
	int ret;
	const struct prop_group_spec *group;

	static char *kwlist[] = {"name", "fallback", "timeout", NULL};
	const char *name;
	uint32_t timeout = self->timeout;
	int fallback = false;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|pI", kwlist, &name, &fallback, &timeout))
		return NULL;

	group = sl_prop_group_spec_find_by_name(self->spec, name);
	if (!group) {
		PyErr_Format(PyExc_ValueError, "No such property group \"%s\"", name);
		return NULL;
	}

	Py_BEGIN_ALLOW_THREADS
	ret = sl_prop_remote_load(self->node, self->port, timeout, group->id, fallback);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		PyErr_Format(PyExc_Exception, "Failed to load group, error %d", ret);
		return NULL;
	}

	Py_RETURN_NONE;
}

static PyObject *PropNode_erase(PropNodeObject *self, PyObject *args, PyObject *kwds)
{
	int ret;
	const struct prop_group_spec *group;

	static char *kwlist[] = {"name", "fallback", "timeout", NULL};
	const char *name;
	uint32_t timeout = self->timeout;
	int fallback = false;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|pI", kwlist, &name, &fallback, &timeout))
		return NULL;

	group = sl_prop_group_spec_find_by_name(self->spec, name);
	if (!group) {
		PyErr_Format(PyExc_ValueError, "No such property group \"%s\"", name);
		return NULL;
	}

	Py_BEGIN_ALLOW_THREADS
	ret = sl_prop_remote_erase(self->node, self->port, timeout, group->id, fallback);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		PyErr_Format(PyExc_Exception, "Failed to erase group, error %d", ret);
		return NULL;
	}

	Py_RETURN_NONE;
}

static PyObject *PropNode_lock(PropNodeObject *self, PyObject *args, PyObject *kwds)
{
	int ret;
	const struct prop_group_spec *group;

	static char *kwlist[] = {"name", "fallback", "timeout", NULL};
	const char *name;
	uint32_t timeout = self->timeout;
	int fallback = false;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|pI", kwlist, &name, &fallback, &timeout))
		return NULL;

	group = sl_prop_group_spec_find_by_name(self->spec, name);
	if (!group) {
		PyErr_Format(PyExc_ValueError, "No such property group \"%s\"", name);
		return NULL;
	}

	Py_BEGIN_ALLOW_THREADS
	ret = sl_prop_remote_lock(self->node, self->port, timeout, group->id, fallback);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		PyErr_Format(PyExc_Exception, "Failed to lock group, error %d", ret);
		return NULL;
	}

	Py_RETURN_NONE;
}

static PyObject *PropNode_unlock(PropNodeObject *self, PyObject *args, PyObject *kwds)
{
	int ret;
	const struct prop_group_spec *group;

	static char *kwlist[] = {"name", "fallback", "timeout", NULL};
	const char *name;
	uint32_t timeout = self->timeout;
	int fallback = false;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|pI", kwlist, &name, &fallback, &timeout))
		return NULL;

	group = sl_prop_group_spec_find_by_name(self->spec, name);
	if (!group) {
		PyErr_Format(PyExc_ValueError, "No such property group \"%s\"", name);
		return NULL;
	}

	Py_BEGIN_ALLOW_THREADS
	ret = sl_prop_remote_unlock(self->node, self->port, timeout, group->id, fallback);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		PyErr_Format(PyExc_Exception, "Failed to unlock group, error %d", ret);
		return NULL;
	}

	Py_RETURN_NONE;
}

static PyMethodDef PropNode_methods[] = {
	{"get",    (PyCFunction)PropNode_get,    METH_VARARGS | METH_KEYWORDS, "Get property value."},
	{"set",    (PyCFunction)PropNode_set,    METH_VARARGS | METH_KEYWORDS, "Set property value."},
	{"save",   (PyCFunction)PropNode_save,   METH_VARARGS | METH_KEYWORDS, "Save property group."},
	{"load",   (PyCFunction)PropNode_load,   METH_VARARGS | METH_KEYWORDS, "Load property group."},
	{"erase",  (PyCFunction)PropNode_erase,  METH_VARARGS | METH_KEYWORDS, "Erase property group."},
	{"lock",   (PyCFunction)PropNode_lock,   METH_VARARGS | METH_KEYWORDS, "Lock property group."},
	{"unlock", (PyCFunction)PropNode_unlock, METH_VARARGS | METH_KEYWORDS, "Unlock property group."},
	{NULL}     /* Sentinel */
};

PyObject *PropNode_mapping_subscript(PropNodeObject *self, PyObject *key)
{
	return PropNode_get_ex(self, key, self->timeout);
}

int PropNode_mapping_ass_subscript(PropNodeObject *self, PyObject *key, PyObject *value)
{
	/* Deletion is not supported */
	if (!value) {
		PyErr_Format(PyExc_ValueError, "Deletion is not supported");
		return -1;
	}

	PyObject *ret = PropNode_set_ex(self, key, value, self->timeout) ;
	if (!ret)
		return -1;

	/* Decrement reference to None object */
	Py_DECREF(ret);
	return 0;
}

static PyMappingMethods PropNode_mapping_methods = {
	.mp_subscript = (binaryfunc)PropNode_mapping_subscript,
	.mp_ass_subscript = (objobjargproc)PropNode_mapping_ass_subscript,
};

static PyTypeObject PropNodeType = {
	PyVarObject_HEAD_INIT(NULL, 0)
	.tp_name = "propclient.PropNode",
	.tp_doc = "PropNode object",
	.tp_basicsize = sizeof(PropNodeObject),
	.tp_itemsize = 0,
	.tp_flags = Py_TPFLAGS_DEFAULT,
	.tp_new = PyType_GenericNew,
	.tp_init = (initproc) PropNode_init,
	.tp_methods = PropNode_methods,
	.tp_as_mapping = &PropNode_mapping_methods,
};

/* Module definition */
static struct PyModuleDef propclient_module = {
	.m_base = PyModuleDef_HEAD_INIT,
	.m_name = "propclient",
	.m_size = -1,
};

PyMODINIT_FUNC PyInit_propclient_python(void)
{
	if (PyType_Ready(&PropNodeType) < 0)
		return NULL;

	/* Create module */
	PyObject *m = PyModule_Create(&propclient_module);
	if (!m)
		return NULL;

	/* Add PropNode object */
	Py_INCREF(&PropNodeType);
	if (PyModule_AddObject(m, "PropNode", (PyObject *) &PropNodeType) < 0) {
		Py_DECREF(&PropNodeType);
		Py_DECREF(m);
		return NULL;
	}

	return m;
}
